/*      X Window Video Capture Testing Tool
 *
 *      For testing a Video for Linux Two video capture driver
 *
 *      This program was written by Olivier Carmona, based on Bill Dirks.
 *      This program is in the public domain.
 *
 */

#include <unistd.h>
#include <stdlib.h>
#include <sys/time.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <fcntl.h>
#include <stdio.h>
#include <sys/mman.h>
#include <errno.h>

#include <linux/fs.h>
#include <linux/kernel.h>
#include <linux/videodev.h>

#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Xaw/Box.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Porthole.h>
#include <X11/Xaw/MenuButton.h>
#include <X11/Xaw/SimpleMenu.h>
#include <X11/Xaw/SmeBSB.h>

/*  This program does not properly detect when the X Windows desktop is
 *  32 bits per pixel. (Some X functions use a value of 24 for the depth
 *  when it is really 32 for reasons I don't understand.) As a workaround,
 *  if you are running a 32-bpp X Windows screen, use -b on the command
 *  line, or uncomment this line:  
#define X_DEPTH 32
*/

/*  Set these according to your capture device set-up */
#define VID_DEVICE 	"/dev/video0" /* Device Name */
#define VID_PIX_FLAGS 	V4L2_FMT_FLAG_ODDFIELD /* Video Output Format */
#define VID_STD	  	"NTSC" /* Default Standard */
#define VID_WIDTH  	320 /* Default Width */
#define VID_HEIGHT 	240 /* Default Height */
#define VID_STREAMING 	1 /* Streaming mode boolean */
#define STREAMBUFS	4 /* Number of streaming buffer */

/* Device Catpure Objects */
typedef struct tag_vimage
  {
    struct v4l2_buffer vidbuf;
    char *data;
    int   length;
  }
VIMAGE;

typedef struct
{
  int fd_video;
  int width;
  int height;
  int depth;
  int pixelformat;
  VIMAGE vimage[STREAMBUFS];
  struct v4l2_format fmt;
  struct v4l2_requestbuffers req;
}capture_device;

/* Windows Display Objects */
typedef struct capture_window
{
  XtAppContext *app;
  Widget       w_toplevel;
  Widget       w_topbox;
  Widget       w_menubox;
  Widget       w_quit;
  Widget       w_viewport;
  Widget       w_stdbutton;
  Widget       w_stdmenu;
  Widget       w_sizebutton;
  Widget       w_sizemenu;
  GC           imagegc;
  int          x_depth;
}capture_window;

/* Standard Menu Objects */
typedef struct 
{
  Widget w_item;
  struct v4l2_standard std;
  capture_window *capt_w;
  capture_device *capt_d;
}
StdMenuItem;

/* Size Menu Objects */
struct
  {
    int width, height;
  }
capturesize[] =
{
  { 160, 120 } ,
  { 192, 144 } ,
  { 320, 240 } ,
  { 384, 288 } ,
  { 640, 480 } ,
  { 768, 576 } ,
};
 
typedef struct
{
  int index;
  capture_window *capt_w;
  capture_device *capt_d;
}
SzMenuItem;

XImage *capt_ximg[STREAMBUFS];


/***************************/
/* Device Catpure Routines */
/***************************/

int capture_fmt (int x_depth)
{
  int fbf = V4L2_PIX_FMT_RGB565;

  if (x_depth == 15)
    fbf = V4L2_PIX_FMT_RGB555;
  else if (x_depth == 16)
    fbf = V4L2_PIX_FMT_RGB565;
  else if (x_depth == 24)
    fbf = V4L2_PIX_FMT_BGR24;
  else if (x_depth == 32)
    fbf = V4L2_PIX_FMT_BGR32;
  else
    {
      printf ("Unrecognized display depth. You may get an "
	      "X Windows error.\n");
    }
    
  return fbf;
}


int
capture_perf (capture_device * pcapt_d)
{
  struct v4l2_performance perf;
 
  if (ioctl (pcapt_d->fd_video, VIDIOC_G_PERF, &perf) != 0)
    printf ("G_PERF returned error code %d\n", errno);
  else
    printf ("Captured %d frames and lost %d frames\n", perf.frames,
	    perf.framesdropped);
  return (perf.frames);
}

void
capture_end (capture_device * pcapt_d)
{
  close (pcapt_d->fd_video);
}

void
capture_std (capture_device * pcapt_d, struct v4l2_standard *std)
{
 int err;
 
// v4l2_std NTSC to Standard
 err = ioctl (pcapt_d->fd_video, VIDIOC_S_STD, std);
 if (err)
   {
     perror ("S_STD in capture_std");
   }
}

void
capture_size (capture_device * pcapt_d, int width, int height)
{
 int err;

 pcapt_d->fmt.fmt.pix.width = width;
 pcapt_d->fmt.fmt.pix.height = height;
 err = ioctl (pcapt_d->fd_video, VIDIOC_S_FMT, &pcapt_d->fmt);
 if (err)
   {
     perror ("S_FMT in capture_size");
   }
}

void
display_free (XImage *XImgTmp)
{
  if (XImgTmp)
    XDestroyImage(XImgTmp);
   
  return;
}

XImage *
display_alloc (capture_window * pcapt_w, char *data, int width, int height)
{
  XImage *XImgTmp;

  XImgTmp = XCreateImage (XtDisplay (pcapt_w->w_viewport),
 	       DefaultVisual (XtDisplay (pcapt_w->w_viewport), 0),
	       (pcapt_w->x_depth == 32) ? 24 : pcapt_w->x_depth,
	       ZPixmap, 0, data, width, height, 16, 0);
	       
  if (XImgTmp == NULL)
    {
      printf ("No XImage\n");
    }
  
  return XImgTmp;
}

void
capture_stop (capture_device * pcapt_d, int streaming)
{
  int i, err;

  if (pcapt_d->fd_video >= 0)
    {
      if (streaming)
        {
          i = V4L2_BUF_TYPE_CAPTURE;
          err = ioctl (pcapt_d->fd_video, VIDIOC_STREAMOFF, &i);
          for (i = 0; i < pcapt_d->req.count; ++i)
  	    {
	      if (pcapt_d->vimage[i].data)
	        munmap (pcapt_d->vimage[i].data, pcapt_d->vimage[i].length);
	      pcapt_d->vimage[i].data = NULL;
	      capt_ximg[i]->data = NULL;
	      display_free (capt_ximg[i]);
	    }
	}
      else
        {
          if (pcapt_d->vimage[0].data)
  	    free (pcapt_d->vimage[0].data);
          pcapt_d->vimage[0].data = NULL;
	  capt_ximg[i]->data = NULL;
	  display_free (capt_ximg[0]);
	}
    }
}

int
capture_start (capture_device * pcapt_d, int streaming)
{
  struct v4l2_standard std;
  int i, err;

  /* Get Video Standard */
  err = ioctl (pcapt_d->fd_video, VIDIOC_G_STD, &std);
  if (err)
    perror ("G_STD in capture_start");

  /* Get Video Format */
  pcapt_d->fmt.type = V4L2_BUF_TYPE_CAPTURE;
  err = ioctl (pcapt_d->fd_video, VIDIOC_G_FMT, &pcapt_d->fmt);
  if (err)
    perror ("G_FMT in capture_start");

  pcapt_d->width = pcapt_d->fmt.fmt.pix.width;
  pcapt_d->height = pcapt_d->fmt.fmt.pix.height;
  pcapt_d->depth = pcapt_d->fmt.fmt.pix.depth;
  pcapt_d->pixelformat = pcapt_d->fmt.fmt.pix.pixelformat;
  
 if (streaming)
   {
     /* Ask Video Device for Buffers */
     pcapt_d->req.count = STREAMBUFS;
     pcapt_d->req.type = V4L2_BUF_TYPE_CAPTURE;
     err = ioctl (pcapt_d->fd_video, VIDIOC_REQBUFS, &pcapt_d->req);
     if (err < 0 || pcapt_d->req.count < 1)
       {
         perror ("REQBUFS in capture _start\n");
         return 0;
       }
    
     /* Query each buffer and map it to the video device */
     for (i = 0; i < pcapt_d->req.count; ++i)
       {
         struct v4l2_buffer *vidbuf = &pcapt_d->vimage[i].vidbuf;
      
         vidbuf->index = i;
         vidbuf->type = V4L2_BUF_TYPE_CAPTURE;
         err = ioctl (pcapt_d->fd_video, VIDIOC_QUERYBUF, vidbuf);
         if (err < 0)
   	   {
	     perror ("QUERYBUF in capture_start");
	     return 0;
	   }
	   
         pcapt_d->vimage[i].length = 0;
         pcapt_d->vimage[i].data = mmap (0, vidbuf->length, 
	           PROT_READ|PROT_WRITE, MAP_SHARED,
	           pcapt_d->fd_video, vidbuf->offset);
         if ((int) pcapt_d->vimage[i].data == -1)
	   {
	     perror ("mmap() in capture_start");
	     return 0;
   	   }
         pcapt_d->vimage[i].length = vidbuf->length; 

	 err = ioctl (pcapt_d->fd_video, VIDIOC_QBUF, vidbuf);
         if (err)
           {
	     perror ("QBUF in capture_start");
	     return 0;
           }
       }
       
     /* Set video stream capture on */
     err = ioctl (pcapt_d->fd_video, VIDIOC_STREAMON, &pcapt_d->vimage[0].vidbuf.type);
     if (err)
       perror ("STREAMON in capture_start");
   }
 else
    {
      /* Alloc one Buffer for one image */
      int sizeimg = pcapt_d->width*pcapt_d->height*pcapt_d->depth/8;
      
      pcapt_d->req.count = 1;
      pcapt_d->vimage[0].data = malloc (sizeimg);
      if (pcapt_d->vimage[0].data == NULL)
        {
          printf ("malloc(%d) failed in capture_start\n", sizeimg);
          return 0;
        }
    }

  return 1;
}


int
capture_init (capture_device * pcapt_d, char *my_device, 
              unsigned int format, int width, int height, int depth)
{
  int err;
  struct v4l2_capability cap;
  struct v4l2_streamparm parm;
  struct v4l2_standard std;

  /* Open Video Device */
  pcapt_d->fd_video = open (my_device, O_RDWR);
  if (pcapt_d->fd_video < 0)
    {
      printf ("No video device \"%s\"\n", my_device);
      return 0;
    }

  /* Querry Video Device Capabilities */
  err = ioctl (pcapt_d->fd_video, VIDIOC_QUERYCAP, &cap);
  if (err)
    {
      perror ("QUERYCAP in capture_init");
      return 0;
    }
  if (cap.type != V4L2_TYPE_CAPTURE)
    {
      printf ("Not a capture device.\n");
      return 0;
    }
  if (!(cap.flags & V4L2_FLAG_READ))
    {
      printf ("Device does not support the read() call.\n");
    }
  if (!(cap.flags & V4L2_FLAG_STREAMING))
    {
      printf ("Device does not support streaming capture.\n");
    }


  /* Set Video Parameters */
  parm.type = V4L2_BUF_TYPE_CAPTURE;
  err = ioctl (pcapt_d->fd_video, VIDIOC_G_PARM, &parm);
  if (err)
    perror ("G_PARM in capture_init");
  
  err = ioctl (pcapt_d->fd_video, VIDIOC_S_PARM, &parm);
  if (err)
    perror ("S_PARM in capture_init");
    
  /* Get Video Standard */
  err = ioctl (pcapt_d->fd_video, VIDIOC_G_STD, &std);
  if (err)
    perror ("G_STD in capture_init");

  /* Set Video Format */
  pcapt_d->fmt.type = V4L2_BUF_TYPE_CAPTURE;
  err = ioctl (pcapt_d->fd_video, VIDIOC_G_FMT, &pcapt_d->fmt);
  if (err)
    perror ("G_FMT in capture_init");

  if (width)  
    pcapt_d->fmt.fmt.pix.width = width;
  else
    pcapt_d->fmt.fmt.pix.width = VID_WIDTH;
  if (height) 
    pcapt_d->fmt.fmt.pix.height = height;
  else
    pcapt_d->fmt.fmt.pix.height = VID_HEIGHT;
  if (depth) 
    pcapt_d->fmt.fmt.pix.depth = depth;
    
  pcapt_d->fmt.fmt.pix.flags |= VID_PIX_FLAGS;
#ifdef VID_PIXELFORMAT
  pcapt_d->fmt.fmt.pix.pixelformat = VID_PIXELFORMAT;
#else
  pcapt_d->fmt.fmt.pix.pixelformat = format;
#endif

  pcapt_d->width = pcapt_d->fmt.fmt.pix.width;
  pcapt_d->height = pcapt_d->fmt.fmt.pix.height;
  pcapt_d->depth = pcapt_d->fmt.fmt.pix.depth;
  pcapt_d->pixelformat = pcapt_d->fmt.fmt.pix.pixelformat;
  
  err = ioctl (pcapt_d->fd_video, VIDIOC_S_FMT, &pcapt_d->fmt);
  if (err)
    perror ("S_FMT in capture_init");
  
  return (1);
}

int
capture_read (capture_device * pcapt_d, int sz_img)
{
  int err;
  
  err = read (pcapt_d->fd_video, pcapt_d->vimage[0].data, sz_img);
  if (err < 0)
    {
      perror ("READ in capture_read");
      return 0;
    }
  
  return 1;
}

int
capture_dq (capture_device * pcapt_d,
	    struct v4l2_buffer *pbuf,
	    int streaming)
{
  int err, n;
  fd_set rdset;
  struct timeval timeout;

  if (streaming)
    {
      FD_ZERO (&rdset);
      FD_SET (pcapt_d->fd_video, &rdset);

      timeout.tv_sec = 1;
      timeout.tv_usec = 0;

      n = select (pcapt_d->fd_video + 1, &rdset, NULL, NULL, &timeout);
      if (n == -1)
	fprintf (stderr, "select error.\n");
      else if (n == 0)
	fprintf (stderr, "select timeout\n");
      else if (FD_ISSET (pcapt_d->fd_video, &rdset))
	{
	  pbuf->type = pcapt_d->vimage[0].vidbuf.type;
	  err = ioctl (pcapt_d->fd_video, VIDIOC_DQBUF, pbuf);
	  if (err)
	    {
	      perror ("DQBUF in capture_dq");
	      return 0;
	    }
	  return 1;
	}
    }
  else
    {
      capture_read(pcapt_d, 1000000);
      pbuf->index = 0;
      return 1;
    }
  return 0;
}

void
capture_q (capture_device * pcapt_d,
	   struct v4l2_buffer *pbuf,
	   int streaming)
{
  int err;

  if (streaming)
    {
      err = ioctl (pcapt_d->fd_video, VIDIOC_QBUF, pbuf);
      if (err)
	perror ("QBUF in capture_q");
    }
}


/****************************/
/* Windows Display Routines */
/****************************/

int display_get_depth (capture_window * pcapt_w)
{
  if (pcapt_w->x_depth == 0)
    {
#ifdef X_DEPTH
      pcapt_w->x_depth = X_DEPTH;
#else
      Display *d = XtDisplay (pcapt_w->w_toplevel);
      Window w = DefaultRootWindow (XtDisplay (pcapt_w->w_toplevel));

      XWindowAttributes wa;
      XGetWindowAttributes (d, w, &wa);
      pcapt_w->x_depth = wa.depth;
#endif
    }
  return (pcapt_w->x_depth);
}

int
display_start (capture_device * pcapt_d, 
	       capture_window * pcapt_w, int width, int height)
{
  Arg 	arg[2];
  int	i;

  for (i=0; i < pcapt_d->req.count; i++)
    {
      capt_ximg[i] = display_alloc (pcapt_w, pcapt_d->vimage[i].data, 
                                    pcapt_d->width, pcapt_d->height);
    }
  
  XtSetArg (arg[0], XtNwidth, width);
  XtSetArg (arg[1], XtNheight, height);
  XtSetValues (pcapt_w->w_viewport, arg, 2);
  XtRealizeWidget (pcapt_w->w_toplevel);

  return 1;
}

void
display_stop (capture_device *pcapt_d, capture_window * pcapt_w)
{
  return;
}

void
quit_callback (Widget w, XtPointer client, XtPointer call)
{
  capture_device *pcapt_d = (capture_device *) client;

  if (pcapt_d->fd_video >= 0)
    {
      capture_perf(pcapt_d);
      capture_stop (pcapt_d, VID_STREAMING);
      capture_end (pcapt_d); 
    }
  exit (0);
}

void
std_menu_callback (Widget w, XtPointer client, XtPointer call)
{
  StdMenuItem *smi = (StdMenuItem *) client;
  capture_device *pcapt_d = smi->capt_d;
  capture_window *pcapt_w = smi->capt_w;

  capture_stop (pcapt_d, VID_STREAMING);
  display_stop (pcapt_d, pcapt_w);
  capture_std (pcapt_d, &smi->std);
  capture_start (pcapt_d, VID_STREAMING);
  display_start (pcapt_d, pcapt_w, pcapt_d->width, pcapt_d->height);
}

void
size_menu_callback (Widget w, XtPointer client, XtPointer call)
{
  SzMenuItem *szi = (SzMenuItem *) client;
  int i = szi->index;
  capture_window *pcapt_w = szi->capt_w;
  capture_device *pcapt_d = szi->capt_d;
  
  capture_stop (pcapt_d, VID_STREAMING);
  display_stop (pcapt_d, pcapt_w);
  capture_size (pcapt_d, capturesize[i].width, capturesize[i].height);
  capture_start (pcapt_d, VID_STREAMING);
  display_start (pcapt_d, pcapt_w, pcapt_d->width, pcapt_d->height);
}

int
display_init (capture_window * pcapt_w, 
              capture_device * pcapt_d,
	      char *my_device,
	      int argc, char *argv[])
{
  int err, i;
  int width, height, format;

  pcapt_w->app = (XtAppContext *) malloc (sizeof (XtAppContext));
  pcapt_w->w_toplevel = XtAppInitialize (pcapt_w->app, "DTSS", NULL, 0,
					 &argc, argv, NULL, NULL, 0);

  display_get_depth (pcapt_w);
  format = capture_fmt(pcapt_w->x_depth);
  capture_init (pcapt_d, my_device, format, 0, 0, 0);
  width = pcapt_d->fmt.fmt.pix.width;
  height = pcapt_d->fmt.fmt.pix.height;
  
  pcapt_w->w_topbox = XtVaCreateManagedWidget ("topbox",
					       boxWidgetClass,
					       pcapt_w->w_toplevel,
					       XtNorientation,
					       XtorientVertical,
					       XtNborderWidth, 0, XtNhSpace,
					       0, XtNvSpace, 0, NULL);
  pcapt_w->w_menubox =
    XtVaCreateManagedWidget ("menubox", boxWidgetClass, pcapt_w->w_topbox,
			     XtNorientation, XtorientHorizontal,
			     XtNborderWidth, 0, XtNhSpace, 4, XtNvSpace, 1,
			     NULL);

  pcapt_w->w_stdbutton = XtVaCreateManagedWidget ("menubutton",
						  menuButtonWidgetClass,
						  pcapt_w->w_menubox,
						  XtNlabel, "Standard",
						  XtNmenuName, "standard",
						  XtNborderWidth, 0, NULL);
  pcapt_w->w_stdmenu = XtVaCreatePopupShell ("standard",
					     simpleMenuWidgetClass,
					     pcapt_w->w_stdbutton, NULL);


  for (i = 0;; ++i)
    {
      struct v4l2_enumstd estd;
      StdMenuItem *smi;

      estd.index = i;
      err = ioctl (pcapt_d->fd_video, VIDIOC_ENUMSTD, &estd);
      if (err)
	break;
      smi = (StdMenuItem *) malloc (sizeof (StdMenuItem));
      smi->std = estd.std;
      smi->capt_w = pcapt_w;
      smi->capt_d = pcapt_d;
      smi->w_item = XtVaCreateManagedWidget (estd.std.name,
					     smeBSBObjectClass,
					     pcapt_w->w_stdmenu, XtNlabel,
					     estd.std.name, NULL);
      XtAddCallback (smi->w_item, XtNcallback, std_menu_callback, smi);
    }

  pcapt_w->w_sizebutton = XtVaCreateManagedWidget ("menubutton",
						   menuButtonWidgetClass,
						   pcapt_w->w_menubox,
						   XtNlabel, "Size",
						   XtNmenuName, "size",
						   XtNborderWidth, 0, NULL);
  pcapt_w->w_sizemenu = XtVaCreatePopupShell ("size",
					      simpleMenuWidgetClass,
					      pcapt_w->w_sizebutton, NULL);
  for (i = 0; i < sizeof (capturesize) / sizeof (capturesize[0]); ++i)
    {
      char sizestr[32];
      SzMenuItem *szi;
      Widget w;

      szi = (SzMenuItem *) malloc (sizeof (SzMenuItem));
      szi->index = i;
      szi->capt_w = pcapt_w;
      szi->capt_d = pcapt_d;
      sprintf (sizestr, "%d x %d", capturesize[i].width,
	       capturesize[i].height);
      w = XtVaCreateManagedWidget (sizestr,
				   smeBSBObjectClass, pcapt_w->w_sizemenu,
				   XtNlabel, sizestr, NULL);

      XtAddCallback (w, XtNcallback, size_menu_callback, (char *) szi);
    }

  pcapt_w->w_quit = XtVaCreateManagedWidget ("Quit!",
					     commandWidgetClass,
					     pcapt_w->w_menubox,
					     XtNborderWidth, 0, NULL);
  XtAddCallback (pcapt_w->w_quit, XtNcallback, quit_callback, pcapt_d);

  pcapt_w->w_viewport = XtVaCreateManagedWidget ("port",
						 portholeWidgetClass,
						 pcapt_w->w_topbox, XtNwidth,
						 width, XtNheight,
						 height, NULL);
  XtRealizeWidget (pcapt_w->w_toplevel);
  pcapt_w->imagegc = XCreateGC (XtDisplay (pcapt_w->w_viewport),
				XtWindow (pcapt_w->w_viewport), 0, NULL);
  if (pcapt_w->imagegc == NULL)
    {
      printf ("XCreateGC failed.\n");
      return 0;
    }

  return 1;
}

int
display_status (capture_window * pcapt_w)
{
  XEvent event;

  if (XtIsRealized (pcapt_w->w_viewport))
    {
      return 1;
    }
  else
    {
      XtAppPeekEvent (*pcapt_w->app, &event);
      return 0;
    }

}

void
display_refresh (capture_window * pcapt_w, 
                 XImage *XImgTmp, int width, int height)
{

  XPutImage (XtDisplay (pcapt_w->w_viewport),
	     XtWindow (pcapt_w->w_viewport),
	     pcapt_w->imagegc, XImgTmp, 0, 0, 0, 0, width, height);
}


void
display_wakeup (capture_window * pcapt_w)
{
  XEvent event;

  while (XtAppPending (*pcapt_w->app))
    {
      XtAppNextEvent (*pcapt_w->app, &event);
      switch (event.type)
	{
	  default:
	    break;
	}
      XtDispatchEvent (&event);
    }
}

void grabSetFps(int fd, int fps)
{
  struct v4l2_streamparm params;

  printf("called v4l2_set_fps with fps=%d\n",fps);
  params.type = V4L2_BUF_TYPE_CAPTURE;
  ioctl(fd, VIDIOC_G_PARM, &params);
  printf("time per frame is: %ld\n", params.parm.capture.timeperframe);
  params.parm.capture.capturemode |= V4L2_CAP_TIMEPERFRAME;
  params.parm.capture.timeperframe = 10000000 / fps;
  if (fps == 30)
    params.parm.capture.timeperframe = 333667;
  printf("time per frame is: %ld\n", params.parm.capture.timeperframe);
  ioctl(fd, VIDIOC_S_PARM, &params);
  
  params.parm.capture.timeperframe = 0;
  ioctl(fd, VIDIOC_G_PARM, &params);
  printf("time per frame is: %ld\n", params.parm.capture.timeperframe);  
}

int
main (int argc, char *argv[])
{
  capture_window capt_w;
  capture_device capt_d;
  int argi;

  /* Video Parameters */
  char my_device[64];

  /*    Put in the device node name */
  strcpy (my_device, VID_DEVICE);

  /* Read arguments */
  capt_w.x_depth = 0; /* detect */
  for (argi = 1; argi < argc; ++argi)
    {
      if (strcmp (argv[argi], "-b") == 0)
	{
	  capt_w.x_depth = 32;
	  continue;
	}
      strcpy (my_device, argv[argi]);
    }

  display_init (&capt_w, &capt_d, my_device, argc, argv);
  
  grabSetFps(capt_d.fd_video, 30);

  if (!capture_start (&capt_d, VID_STREAMING))
    exit(1);
  
  display_start (&capt_d, &capt_w, capt_d.width, capt_d.height);
    
  printf ("Capturing %dx%dx%d \"%4.4s\" images\n",
	  capt_d.width, capt_d.height, capt_d.depth,
	  (char *) &(capt_d.pixelformat));
  printf ("Images are %d bytes each\n", 
          capt_d.width * capt_d.height * capt_d.depth / 8);

  for (;;)
    {
      struct v4l2_buffer tempbuf;

      if (display_status (&capt_w))
	{
	  if (capture_dq (&capt_d, &tempbuf, VID_STREAMING))
	    {
	      display_refresh (&capt_w, capt_ximg[tempbuf.index], 
			       capt_d.width, capt_d.height);
  	      capture_q (&capt_d, &tempbuf, VID_STREAMING);
	    }
	}

      display_wakeup (&capt_w);
    }

  return 0;
}

